@@ -24,6 +24,7 @@ import (
 	"github.com/openshift/cluster-version-operator/lib"
 	"github.com/openshift/cluster-version-operator/lib/resourcebuilder"
 	"github.com/openshift/cluster-version-operator/pkg/payload"
+	"github.com/openshift/cluster-version-operator/pkg/payload/precondition"
 )
 
 // ConfigSyncWorker abstracts how the image is synchronized to the server. Introduced for testing.
@@ -125,10 +126,11 @@ func (w SyncWorkerStatus) DeepCopy() *SyncWorkerStatus {
 //     syncOnce() returns nil -> Reconciling
 //
 type SyncWorker struct {
-	backoff     wait.Backoff
-	retriever   PayloadRetriever
-	builder     payload.ResourceBuilder
-	reconciling bool
+	backoff       wait.Backoff
+	retriever     PayloadRetriever
+	builder       payload.ResourceBuilder
+	preconditions precondition.List
+	reconciling   bool
 
 	// minimumReconcileInterval is the minimum time between reconcile attempts, and is
 	// used to define the maximum backoff interval when syncOnce() returns an error.
@@ -146,11 +148,16 @@ type SyncWorker struct {
 
 	// updated by the run method only
 	payload *payload.Update
+
+	// exclude is an identifier used to determine which
+	// manifests should be excluded based on an annotation
+	// of the form exclude.release.openshift.io/<identifier>=true
+	exclude string
 }
 
 // NewSyncWorker initializes a ConfigSyncWorker that will retrieve payloads to disk, apply them via builder
 // to a server, and obey limits about how often to reconcile or retry on errors.
-func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder, reconcileInterval time.Duration, backoff wait.Backoff) ConfigSyncWorker {
+func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder, reconcileInterval time.Duration, backoff wait.Backoff, exclude string) *SyncWorker {
 	return &SyncWorker{
 		retriever: retriever,
 		builder:   builder,
@@ -163,9 +170,20 @@ func NewSyncWorker(retriever PayloadRetriever, builder payload.ResourceBuilder,
 		// Status() or use the result of calling Update() instead because the channel can be out of date
 		// if the reader is not fast enough.
 		report: make(chan SyncWorkerStatus, 500),
+
+		exclude: exclude,
 	}
 }
 
+// NewSyncWorkerWithPreconditions initializes a ConfigSyncWorker that will retrieve payloads to disk, apply them via builder
+// to a server, and obey limits about how often to reconcile or retry on errors.
+// It allows providing preconditions for loading payload.
+func NewSyncWorkerWithPreconditions(retriever PayloadRetriever, builder payload.ResourceBuilder, preconditions precondition.List, reconcileInterval time.Duration, backoff wait.Backoff, exclude string) *SyncWorker {
+	worker := NewSyncWorker(retriever, builder, reconcileInterval, backoff, exclude)
+	worker.preconditions = preconditions
+	return worker
+}
+
 // StatusCh returns a channel that reports status from the worker. The channel is buffered and events
 // can be lost, so this is best used as a trigger to read the latest status.
 func (w *SyncWorker) StatusCh() <-chan SyncWorkerStatus {
@@ -493,6 +511,30 @@ func (w *SyncWorker) syncOnce(ctx context.Context, work *SyncWork, maxWorkers in
 		payloadUpdate.VerifiedImage = info.Verified
 		payloadUpdate.LoadedAt = time.Now()
 
+		// need to make sure the payload is only set when the preconditions have been successfull
+		if !info.Local && len(w.preconditions) > 0 {
+			reporter.Report(SyncWorkerStatus{
+				Generation:  work.Generation,
+				Step:        "PreconditionChecks",
+				Initial:     work.State.Initializing(),
+				Reconciling: work.State.Reconciling(),
+				Actual:      update,
+				Verified:    info.Verified,
+			})
+			if err := precondition.Summarize(w.preconditions.RunAll(ctx)); err != nil && !update.Force {
+				reporter.Report(SyncWorkerStatus{
+					Generation:  work.Generation,
+					Failure:     err,
+					Step:        "PreconditionChecks",
+					Initial:     work.State.Initializing(),
+					Reconciling: work.State.Reconciling(),
+					Actual:      update,
+					Verified:    info.Verified,
+				})
+				return err
+			}
+		}
+
 		w.payload = payloadUpdate
 		klog.V(4).Infof("Payload loaded from %s with hash %s", payloadUpdate.ReleaseImage, payloadUpdate.ManifestHash)
 	}
@@ -571,15 +613,15 @@ func (w *SyncWorker) apply(ctx context.Context, payloadUpdate *payload.Update, w
 	// update each object
 	errs := payload.RunGraph(ctx, graph, maxWorkers, func(ctx context.Context, tasks []*payload.Task) error {
 		for _, task := range tasks {
-			if err := ctx.Err(); err != nil {
-				return cr.ContextError(err)
+			if contextIsCancelled(ctx) {
+				return cr.CancelError()
 			}
 			cr.Update()
 
 			klog.V(4).Infof("Running sync for %s", task)
 			klog.V(5).Infof("Manifest: %s", string(task.Manifest.Raw))
 
-			ov, ok := getOverrideForManifest(work.Overrides, task.Manifest)
+			ov, ok := getOverrideForManifest(work.Overrides, w.exclude, task.Manifest)
 			if ok && ov.Unmanaged {
 				klog.V(4).Infof("Skipping %s as unmanaged", task)
 				continue
@@ -594,10 +636,8 @@ func (w *SyncWorker) apply(ctx context.Context, payloadUpdate *payload.Update, w
 		return nil
 	})
 	if len(errs) > 0 {
-		if err := cr.Errors(errs); err != nil {
-			return err
-		}
-		return errs[0]
+		err := cr.Errors(errs)
+		return err
 	}
 
 	// update the status
@@ -618,11 +658,11 @@ func init() {
 	)
 }
 
-type errContext struct {
+type errCanceled struct {
 	err error
 }
 
-func (e errContext) Error() string { return e.err.Error() }
+func (e errCanceled) Error() string { return e.err.Error() }
 
 // consistentReporter hides the details of calculating the status based on the progress
 // of the graph runner.
@@ -659,7 +699,7 @@ func (r *consistentReporter) Error(err error) {
 	copied := r.status
 	copied.Step = "ApplyResources"
 	copied.Fraction = float32(r.done) / float32(r.total)
-	if !isContextError(err) {
+	if !isCancelledError(err) {
 		copied.Failure = err
 	}
 	r.reporter.Report(copied)
@@ -680,10 +720,10 @@ func (r *consistentReporter) Errors(errs []error) error {
 	return err
 }
 
-func (r *consistentReporter) ContextError(err error) error {
+func (r *consistentReporter) CancelError() error {
 	r.lock.Lock()
 	defer r.lock.Unlock()
-	return errContext{fmt.Errorf("update %s at %d of %d", err, r.done, r.total)}
+	return errCanceled{fmt.Errorf("update was cancelled at %d of %d", r.done, r.total)}
 }
 
 func (r *consistentReporter) Complete() {
@@ -699,11 +739,11 @@ func (r *consistentReporter) Complete() {
 	r.reporter.Report(copied)
 }
 
-func isContextError(err error) bool {
+func isCancelledError(err error) bool {
 	if err == nil {
 		return false
 	}
-	_, ok := err.(errContext)
+	_, ok := err.(errCanceled)
 	return ok
 }
 
@@ -724,12 +764,11 @@ func isImageVerificationError(err error) bool {
 // not truly an error (cancellation).
 // TODO: take into account install vs upgrade
 func summarizeTaskGraphErrors(errs []error) error {
-	// we ignore context errors (canceled or timed out) since they don't
-	// provide good feedback to users and are an internal detail of the
-	// server
-	err := errors.FilterOut(errors.NewAggregate(errs), isContextError)
+	// we ignore cancellation errors since they don't provide good feedback to users and are an internal
+	// detail of the server
+	err := errors.FilterOut(errors.NewAggregate(errs), isCancelledError)
 	if err == nil {
-		klog.V(4).Infof("All errors were context errors: %v", errs)
+		klog.V(4).Infof("All errors were cancellation errors: %v", errs)
 		return nil
 	}
 	agg, ok := err.(errors.Aggregate)
@@ -874,7 +913,7 @@ func newMultipleError(errs []error) error {
 }
 
 // getOverrideForManifest returns the override and true when override exists for manifest.
-func getOverrideForManifest(overrides []configv1.ComponentOverride, manifest *lib.Manifest) (configv1.ComponentOverride, bool) {
+func getOverrideForManifest(overrides []configv1.ComponentOverride, excludeIdentifier string, manifest *lib.Manifest) (configv1.ComponentOverride, bool) {
 	for idx, ov := range overrides {
 		kind, namespace, name := manifest.GVK.Kind, manifest.Object().GetNamespace(), manifest.Object().GetName()
 		if ov.Kind == kind &&
@@ -883,6 +922,10 @@ func getOverrideForManifest(overrides []configv1.ComponentOverride, manifest *li
 			return overrides[idx], true
 		}
 	}
+	excludeAnnotation := fmt.Sprintf("exclude.release.openshift.io/%s", excludeIdentifier)
+	if annotations := manifest.Object().GetAnnotations(); annotations != nil && annotations[excludeAnnotation] == "true" {
+		return configv1.ComponentOverride{Unmanaged: true}, true
+	}
 	return configv1.ComponentOverride{}, false
 }
 
@@ -896,6 +939,16 @@ func ownerRefModifier(config *configv1.ClusterVersion) resourcebuilder.MetaV1Obj
 	}
 }
 
+// contextIsCancelled returns true if the provided context is cancelled.
+func contextIsCancelled(ctx context.Context) bool {
+	select {
+	case <-ctx.Done():
+		return true
+	default:
+		return false
+	}
+}
+
 // runThrottledStatusNotifier invokes fn every time ch is updated, but no more often than once
 // every interval. If bucket is non-zero then the channel is throttled like a rate limiter bucket.
 func runThrottledStatusNotifier(stopCh <-chan struct{}, interval time.Duration, bucket int, ch <-chan SyncWorkerStatus, fn func()) {
